# This module defines utilities for the cloud function api

from .utils import __API_VERSION__, generate_default_session
import tempfile
import shutil
import subprocess
import glob
import os
import json
import time

__FUNCTION_MAPPING__ = {
    'create_submission': 'submit.py',
    'abort_submission': 'abort.py',
    'existence': 'internal.py',
    'redacted': 'internal.py',
    'check_abort': 'internal.py',
    'quotas': 'quotas.py',
    'register': 'register.py',
    'query_account': 'query.py',
    'insert_resolution': 'resolution.py',
    'webhook': 'webhook.py',
    'update': 'webhook.py',
    'oauth': 'auth.py'
}

RESOLUTION_URL = "https://us-central1-a-graubert.cloudfunctions.net/resolve-" + __API_VERSION__['resolve']

def __autodeploy_internal(endpoint, namespace):
    from lapdog.gateway import resolve_project_for_namespace
    from lapdog.cloud.patch import __ENDPOINTS__ as endpoint_map
    project = resolve_project_for_namespace(namespace)
    service_account = 'lapdog-functions@{}.iam.gserviceaccount.com'.format(project)
    function = endpoint_map[endpoint]
    print("Deploying", function, '@', __FUNCTION_MAPPING__[function], '->', endpoint)
    print(service_account, 'in', namespace, '(', project,')')
    input()
    _deploy(function, endpoint, service_account, project)

def _deploy(function, endpoint, service_account=None, project=None, overload_version=None):
    if overload_version is None:
        overload_version = __API_VERSION__[endpoint]
    with tempfile.TemporaryDirectory() as tempdir:
        with open(os.path.join(tempdir, 'requirements.txt'), 'w') as w:
            w.write('google-auth==1.6.3\n')
            w.write('google-cloud-storage==1.20.0\n')
            w.write('google-cloud-kms==1.2.1\n')
            w.write('cryptography==2.7\n')
            w.write('google-cloud-logging==1.12.1\n')
            w.write('google-cloud-core==1.0.3\n')
        shutil.copyfile(
            os.path.join(
                os.path.dirname(__file__),
                __FUNCTION_MAPPING__[function]
            ),
            os.path.join(tempdir, 'main.py')
        )
        shutil.copyfile(
            os.path.join(
                os.path.dirname(__file__),
                'utils.py'
            ),
            os.path.join(tempdir, 'utils.py')
        )
        cmd = 'gcloud {project} functions deploy {endpoint}-{version} --entry-point {function} --runtime python37 --trigger-http --source {path} {service_account}'.format(
            endpoint=endpoint,
            version=overload_version,
            function=function,
            path=tempdir,
            service_account='' if service_account is None else ('--service-account '+service_account),
            project='' if project is None else ('--project '+project)
        )
        print(cmd)
        subprocess.check_call(
            cmd,
            shell=True,
            executable='/bin/bash'
        )
        time.sleep(5) # Give the API a brief chance to catch up to the deployment
        session = generate_default_session()
        policy = session.get(
            'https://cloudfunctions.googleapis.com/v1/projects/{}/locations/us-central1/functions/{}-{}:getIamPolicy'.format(
                project if project is not None else session._default_project,
                endpoint,
                overload_version
            )
        ).json()
        print(policy)
        found = False
        for i, binding in enumerate(policy['bindings']):
            if binding['role'] == 'roles/cloudfunctions.invoker':
                found = True
                policy['bindings'][i]['members'].append('allAuthenticatedUsers')
        if not found:
            policy['bindings'].append({
                'role': 'roles/cloudfunctions.invoker',
                'members': ['allAuthenticatedUsers']
            })
        session.post(
            'https://cloudfunctions.googleapis.com/v1/projects/{}/locations/us-central1/functions/{}-{}:setIamPolicy'.format(
                project if project is not None else session._default_project,
                endpoint,
                overload_version
            ),
            headers={'Content-Type': 'application/json'},
            json={
                'policy': policy,
                'updateMask': 'bindings'
            }
        )

def __generate_alert_internal(title, alert_type, content, text=None):
    """
    You're welcome to use this function, but you won't have permissions
    """
    assert alert_type in {'critical', 'warning', 'info'}
    blob = utils.getblob(
        'gs://lapdog-alerts/{}'.format(title)
    )
    alert = {
        'type': alert_type,
        'content': content
    }
    if text is not None:
        alert['text'] = text
    blob.upload_from_string(json.dumps(alert))



import os
try:
    from . import utils
except ImportError:
    import sys
    sys.path.append(os.path.dirname(__file__))
    import utils
import traceback
import sys
import json
import requests
from hashlib import sha512

@utils.cors("POST")
def oauth(request):
    """
    This function is unique. It is not deployed into each project.
    It is deployed once into my personal project which serves as a centralized
    database.

    This endpoint handles incoming OAuth
    """
    logger = utils.CloudLogger().log_request(request)
    try:
        data = request.get_json()

        if not isinstance(data, dict):
            return (
                {
                    'error': "Bad Request",
                    'message': ("No data was provided" if data is None else "Expected JSON dictionary in request body")
                },
                400
            )

        if 'grant_type' not in data:
            return (
                {
                    'error': 'Missing Parameters',
                    'message': "Missing required 'grant_type' parameter"
                },
                400
            )

        if 'client_id' not in data:
            return (
                {
                    'error': 'Missing Parameters',
                    'message': "Missing required 'client_id' parameter"
                },
                400
            )

        if data['client_id'] != utils.OAUTH_CLIENT_ID:
            return (
                {
                    'error': 'Bad Client',
                    'message': 'The provided client ID did not match the server\'s client ID'
                },
                409
            )

        if data['grant_type'] == 'authorization_code':
            if 'code' not in data:
                return (
                    {
                        'error': 'Missing Parameters',
                        'message': "Missing required 'code' parameter (required by grant_type = authorization_code)"
                    },
                    400
                )
            if 'redirect_uri' not in data:
                return (
                    {
                        'error': 'Missing Parameters',
                        'message': "Missing required 'redirect_uri' parameter (required by grant_type = authorization_code)"
                    },
                    400
                )
            data['client_secret'] = os.environ['OAUTH_CLIENT_SECRET']
            response = requests.post(
                'https://oauth2.googleapis.com/token',
                headers={
                    'Content-Type': 'application/x-www-form-urlencoded'
                },
                data=data
            )
            if response.status_code != 200:
                return (
                    {
                        'error': 'Authorization failed',
                        'message': 'Google rejected authorization request: {}'.format(response.text)
                    },
                    response.status_code
                )
            return response.json(), 200
        elif data['grant_type'] == 'refresh_token':
            if 'refresh_token' not in data:
                return (
                    {
                        'error': 'Missing Parameters',
                        'message': "Missing required 'refresh_token' parameter (required by grant_type = refresh_token)"
                    },
                    400
                )
            data['client_secret'] = os.environ['OAUTH_CLIENT_SECRET']
            response = requests.post(
                'https://oauth2.googleapis.com/token',
                headers={
                    'Content-Type': 'application/x-www-form-urlencoded'
                },
                data=data
            )
            if response.status_code != 200:
                return (
                    {
                        'error': 'Refresh failed',
                        'message': 'Google rejected refresh request: {}'.format(response.text)
                    },
                    response.status_code
                )
            return response.json(), 200
        else:
            return (
                {
                    'error': 'Bad grant_type',
                    'message': 'grant_type must be "authorization_code" or "refresh_token"'
                },
                400
            )
    except:
        logger.log_exception()
        return (
            {
                'error': 'Unknown Error',
                'message': traceback.format_exc()
            },
            500
        )



#!/usr/bin/python

# Copyright 2017 Google Inc.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file or at
# https://developers.google.com/open-source/licenses/bsd

# cromwell_driver.py
#
# This script provides a library interface to Cromwell, namely:
#    * Start the Cromwell server
#    * Submit execution requests to Cromwell
#    * Poll Cromwell for job status

import logging
import os
import subprocess
import time
import json
import csv
from google.cloud import logging as stackdriver
from google.cloud.logging.resource import Resource as LogResource

import requests

import sys_util
import sys
import atexit
import traceback
import itertools
import shlex

def gce_get_metadata(path):
    """Queries the GCE metadata server the specified value."""
    return requests.get(
        'http://metadata/computeMetadata/v1/{}'.format(path),
        headers={
            'Metadata-Flavor': 'Google'
        }
    ).text

def clump(seq, length):
    getter = iter(seq)
    while True:
        try:
            tmp = next(getter)
            yield itertools.chain([tmp], itertools.islice(getter, length-1))
        except StopIteration:
            return

# https://stackoverflow.com/questions/15063936/csv-error-field-larger-than-field-limit-131072
maxInt = sys.maxsize
while True:
    # decrease the maxInt value by factor 10
    # as long as the OverflowError occurs.
    try:
        csv.field_size_limit(maxInt)
        break
    except OverflowError:
        maxInt = int(maxInt/10)

def unpack(data):
    for k,v in data.items():
        if isinstance(v, str):
            if v.startswith('[') and v.endswith(']'):
                try:
                    data[k] = json.loads(v.replace("\\'", '~<BACKQUOTE>').replace("'", '"').replace('~<BACKQUOTE>', "\\'"))
                except:
                    pass
            elif v.startswith('{') and v.endswith('}'):
                try:
                    data[k] = json.loads(v.replace("\\'", '~<BACKQUOTE>').replace("'", '"').replace('~<BACKQUOTE>', "\\'"))
                except:
                    pass
    if len(json.dumps(data)) >= 10485760:
        raise ValueError("The size of input metadata for an individual workflow cannot exceed 10 Mib")
    return data

class CromwellDriver(object):

    def __init__(self, cromwell_conf, cromwell_jar):
        self.cromwell_conf = cromwell_conf
        self.cromwell_jar = cromwell_jar

        self.cromwell_proc = None

        self.batch_submission = False
        self.logger = CloudLogger().log_instance()

    def start(self, memory=3):
        """Start the Cromwell service."""
        if self.cromwell_proc:
            logging.info("Request to start Cromwell: already running")
            return

        self.cromwell_proc = subprocess.Popen([
                'java',
                '-Dconfig.file=' + self.cromwell_conf,
                '-Xmx%dg'%memory,
                '-jar', self.cromwell_jar,
                'server'])

        self.mem = memory

        logging.info("Started Cromwell")
        self.logger.log(
            'Launching cromwell process',
            memory=memory
        )

    def check_cromwell(self):
        status = self.cromwell_proc.poll()
        if status == -9:
            logging.error("Cromwell Proc was killed. Retry with additional memory")
        elif status is not None:
            logging.error("Cromwell Proc exit with status: %d" % status)

    def fetch(self, session, wf_id=None, post=False, files=None, method=None):
        url = 'http://localhost:8000/api/workflows/v1'
        if wf_id is not None:
            url = os.path.join(url, wf_id)
        if method is not None:
            url = os.path.join(url, method)
        if post:
            r = session.post(url, files=files)
        else:
            for attempt in range(3):
                try:
                    r = session.get(url)
                    assert r.status_code < 400
                    break
                except requests.ConnectionError:
                    logging.info("Request failed: Connection Error")
                    if attempt == 2:
                        raise
                    time.sleep(10)
                except AssertionError:
                    logging.info("Request failed (%d) : %s" % (r.status_code, r.text))
                    if attempt == 2:
                        raise
                    time.sleep(10)
                    r.connection.close()
        r.connection.close()
        return r.json()

    def abort(self, workflow_id):
        self.logger.log("Aborting Workflow", workflow_id=workflow_id)
        return requests.post(
            'http://localhost:8000/api/workflows/v1/{0}/abort'.format(workflow_id)
        ).json()

    def batch(self, submission_id, wdl, inputs, options, batch_limit, query_limit):
        logging.info("Starting batch request. Waiting for cromwell to start...")
        self.logger.log(
            "Beginning batch request",
            batch_limit=batch_limit,
            query_limit=query_limit,
        )
        time.sleep(60)
        with open(wdl, 'r') as wdlReader:
            with open(options, 'r') as optionReader:
                opts = json.load(optionReader)
                opts['google_labels'] = {
                    'lapdog-submission-id':'id-'+submission_id,
                    'lapdog-execution-role':'worker'
                }
                data = {
                    'workflowSource': wdlReader.read(),
                    # 'workflowInputs': json.dumps([line for line in reader]),
                    'workflowOptions': json.dumps(opts),
                }
        logging.info("Starting the following configuration: " + json.dumps(data))
        output = []
        first = True
        with open(inputs, 'r') as inputReader:
            reader = csv.DictReader(inputReader, delimiter='\t', lineterminator='\n')
            with requests.Session() as session:
                for batch in clump(reader, batch_limit):
                    self.check_cromwell()
                    logging.info("Running a new batch of %d workflows" % batch_limit)

                    chunk = []

                    if not first:
                        logging.info("Restarting cromwell...")
                        self.cromwell_proc.kill()
                        self.cromwell_proc = None
                        time.sleep(10)
                        self.start(self.mem)
                        time.sleep(20)
                        logging.info("Resuming next batch")
                    else:
                        first = False

                    for group in clump(batch, query_limit):
                        logging.info("Starting a chunk of %d workflows" % query_limit)
                        group = [line for line in group]
                        logging.info("There are %d workflows in this group" % len(group))
                        response = None
                        for attempt in range(10):
                            try:
                                data['workflowInputs'] = json.dumps([unpack(line) for line in group])
                                self.logger.log(
                                    'Launching workflow batch',
                                    json=data
                                )
                                response = session.post(
                                    'http://localhost:8000/api/workflows/v1/batch',
                                    files=data
                                )
                                response = response.json()
                                logging.info("Submitted jobs. Begin polling")
                                break
                            except requests.exceptions.ConnectionError as e:
                                self.logger.log_exception()
                                traceback.print_exc()
                                self.check_cromwell()
                                logging.info("Failed to connect to Cromwell (attempt %d): %s",
                                    attempt + 1, e)
                                time.sleep(30)
                            except ValueError:
                                self.logger.log_exception(
                                    "JSON Decode error",
                                    response=response.text if response is not None else None,
                                )
                                traceback.print_exc()
                                self.check_cromwell()
                                logging.error("Unexpected response from Cromwell: (%d) : %s" % (response.status_code, response.text))
                                raise

                        if not response:
                            self.check_cromwell()
                            self.logging.log(
                                "Cromwell timeout",
                                severity="WARNING"
                            )
                            sys_util.exit_with_error(
                                    "Failed to connect to Cromwell after {0} seconds".format(
                                            300))

                        logging.info("Raw response: " + repr(response))

                        for job in response:
                            if job['status'] != 'Submitted' and job['status'] != 'Running':
                                for job in response:
                                    self.abort(job['id'])
                                self.logging.log(
                                    'Unexpected job status',
                                    status=job['status'],
                                    jobs=response,
                                    severity='ERROR'
                                )
                                sys_util.exit_with_error(
                                        "Job {} status from Cromwell was not 'Submitted', instead '{}'".format(
                                                job['id'], job['status']))
                            else:
                                chunk.append(job)

                    self.batch_submission = True
                    self.check_cromwell()

                    @atexit.register
                    def abort_all_jobs():
                        if self.batch_submission:
                            for job in response:
                                self.abort(job['id'])

                    for i in range(12):
                        time.sleep(5)

                    attempt = 0
                    max_failed_attempts = 3
                    known_failures = set()
                    while True:
                        for i in range(3):
                            time.sleep(10)

                        self.check_cromwell()

                        # Cromwell occassionally fails to respond to the status request.
                        # Only give up after 3 consecutive failed requests.
                        try:
                            status_json = [
                                [
                                    self.fetch(session, wf_id=job['id'], method='status'),
                                    time.sleep(0.1)
                                ][0]
                                for job in chunk
                            ]
                            attempt = 0
                        except requests.exceptions.ConnectionError as e:
                            self.logger.log_exception()
                            attempt += 1
                            logging.info("Error polling Cromwell job status (attempt %d): %s",
                                attempt, e)
                            self.check_cromwell()

                            if attempt >= max_failed_attempts:
                                self.logger.log(
                                    'Cromwell crash with active workflows',
                                    jobs=chunk,
                                    severity='WARNING'
                                )
                                sys_util.exit_with_error(
                                    "Cromwell did not respond for %d consecutive requests" % attempt)

                            continue

                        statuses = {job['status'] for job in status_json}
                        # logging.info("<WORKFLOW STATUS UPDATE> %s" % json.dumps(status_json))
                        if 'Failed' in statuses:
                            new_failures = [
                                job for job in status_json
                                if job['status'] == 'Failed' and job['id'] not in known_failures
                            ]
                            if len(new_failures):
                                sys.stderr.write(
                                     "The following jobs failed: %s\n" % (
                                         ', '.join('%s (%s)' % (job['id'], job['status']) for job in new_failures)
                                     )
                                 )
                            known_failures |= {job['id'] for job in new_failures}
                        if not len(statuses - {'Succeeded', 'Failed', 'Aborted'}):
                            logging.info("All workflows in terminal states")
                            self.logger.log(
                                'Batch complete',
                                json=status_json,
                            )
                            break

                    self.batch_submission = False


                    output += [
                        {
                            'workflow_id':job['id'],
                            'workflow_status':job['status'],
                            'workflow_output': self.fetch(session, wf_id=job['id'], method='outputs') if job['status'] == 'Succeeded' else None,
                            'workflow_metadata': self.fetch(session, wf_id=job['id'], method='metadata') if job['status'] == 'Succeeded' else None,
                        }
                        for job in status_json
                    ]

                    self.check_cromwell()

                    if 'Aborted' in statuses:
                        # Quit now. No reason to start a new batch to get aborted
                        self.logger.log(
                            'Submission aborted',
                            json=output
                        )
                        sys.stderr.write("There were aborted workflows. Aborting submission now.")
                        return output
        logging.info("<SUBMISSION COMPLETE. FINALIZING DATA>")
        self.logger.log(
            'Submission complete. Finalizing data',
            json=output
        )
        return output



    def submit(self, wdl, workflow_inputs, workflow_options, sleep_time=15):
        """Post new job to the server and poll for completion."""

        # Add required input files
        with open(wdl, 'r') as f:
            wdl_source = f.read()
        with open(workflow_inputs, 'r') as f:
            wf_inputs = f.read()

        files = {
                'wdlSource': wdl_source,
                'workflowInputs': wf_inputs,
        }

        # Add workflow options if specified
        if workflow_options:
            with open(workflow_options, 'r') as f:
                wf_options = f.read()
                files['workflowOptions'] = wf_options

        # After Cromwell start, it may take a few seconds to be ready for requests.
        # Poll up to a minute for successful connect and submit.

        job = None
        max_time_wait = 60
        wait_interval = 5

        time.sleep(wait_interval)
        for attempt in range(max_time_wait/wait_interval):
            try:
                job = self.fetch(session, post=True, files=files)
                break
            except requests.exceptions.ConnectionError as e:
                logging.info("Failed to connect to Cromwell (attempt %d): %s",
                    attempt + 1, e)
                time.sleep(wait_interval)

        if not job:
            sys_util.exit_with_error(
                    "Failed to connect to Cromwell after {0} seconds".format(
                            max_time_wait))

        if job['status'] != 'Submitted':
            sys_util.exit_with_error(
                    "Job status from Cromwell was not 'Submitted', instead '{0}'".format(
                            job['status']))

        # Job is running.
        cromwell_id = job['id']
        logging.info("Job submitted to Cromwell. job id: %s", cromwell_id)

        # Poll Cromwell for job completion.
        attempt = 0
        max_failed_attempts = 3
        while True:
            time.sleep(sleep_time)

            # Cromwell occassionally fails to respond to the status request.
            # Only give up after 3 consecutive failed requests.
            try:
                status_json = self.fetch(session, wf_id=cromwell_id, method='status')
                attempt = 0
            except requests.exceptions.ConnectionError as e:
                attempt += 1
                logging.info("Error polling Cromwell job status (attempt %d): %s",
                    attempt, e)

                if attempt >= max_failed_attempts:
                    sys_util.exit_with_error(
                        "Cromwell did not respond for %d consecutive requests" % attempt)

                continue

            status = status_json['status']
            if status == 'Succeeded':
                break
            elif status == 'Submitted':
                pass
            elif status == 'Running':
                pass
            else:
                sys_util.exit_with_error(
                        "Status of job is not Submitted, Running, or Succeeded: %s" % status)

        logging.info("Cromwell job status: %s", status)

        # Cromwell produces a list of outputs and full job details
        outputs = self.fetch(session, wf_id=cromwell_id, method='outputs')
        metadata = self.fetch(session, wf_id=cromwell_id, method='metadata')

        return outputs, metadata


class CloudLogger(object):
    MAX_PAYLOAD_SIZE = 261120
    def __init__(self, project=None, submission_id=None):
        self.logger = stackdriver.Client(project).logger('lapdog-api-logging%2Fcromwell')
        self.labels = {
            'lapdog-entity-type': 'cromell',
            'lapdog-submission-id': submission_id if submission_id is not None else os.environ['LAPDOG_SUBMISSION_ID'],
            'lapdog-engine-project': project if project is not None else os.environ['LAPDOG_PROJECT']
        }

    def log_instance(self):
        self.log(
            'Cromwell Logging started',
            json={
                'instance-type': gce_get_metadata('instance/machine-type'),
                'instance-name': gce_get_metadata('instance/name'),
                'instance-zone': gce_get_metadata('instance/zone'),
                'submission-id': self.labels['lapdog-submission-id'],
                'project': self.labels['lapdog-engine-project']
            },
            severity='DEBUG'
        )
        return self

    def log_exception(self, message='Unhandled Exception', **kwargs):
        self.log(
            message=message,
            traceback=traceback.format_exc(),
            severity='WARNING',
            **kwargs
        )

    def trunate_payload(self, message, labels, severity):
        message = repr(message)
        if len(message) > CloudLogger.MAX_PAYLOAD_SIZE:
            self.logger.log_struct(
                {
                    'message': 'Truncated payload ({} bytes total)'.format(len(message)),
                    'json': {
                        'message_head':  message[:130540],
                        'message_tail': message[-130540:]
                    }
                },
                labels=labels,
                severity=severity
            )
            return True

    def log(self, text=None, json=None, severity='DEFAULT', **kwargs):
        if isinstance(severity, str):
            severity = {
                'DEFAULT': 0,
                'DEBUG': 100,
                'INFO': 200,
                'NOTICE': 300,
                'WARNING': 400,
                'WARN': 400,
                'ERROR': 500,
                'ERR': 500,
                'CRITICAL': 600,
                'ALERT': 700,
                'EMERGENCY': 800
            }[severity]
        if len(kwargs):
            if json is not None:
                json = {k:v for k,v in json.items()}
                json.update(kwargs)
            else:
                json = kwargs
        # Mask user tokens
        if isinstance(json, dict) and 'token' in json:
            json['token'] = '****************'
        if text is None:
            if json is None:
                raise ValueError("No input provided")
            if not self.trunate_payload(json, labels=self.labels, severity=severity):
                self.logger.log_struct(json, labels=self.labels, severity=severity)
        elif json is not None:
            json = {
                'message': text,
                'json': json
            }
            if not self.trunate_payload(json, self.labels, severity):
                self.logger.log_struct(
                    json,
                    labels=self.labels,
                    severity=severity
                )
        elif not self.trunate_payload(text, self.labels, severity):
            self.logger.log_text(
                text,
                labels=self.labels,
                severity=severity
            )


if __name__ == '__main__':
    pass



import os
import base64
import webbrowser
import json
import time
import subprocess
from threading import RLock
from urllib.parse import urlencode
from hashlib import sha256
from getpass import getuser
import requests
import crayons
from google.auth.transport.requests import AuthorizedSession
from  google.oauth2.credentials import Credentials
from .cloud import utils
from .cache import cached

@cached(60, 1)
def get_gcloud_account():
    """
    Gets the currently logged in gcloud account.
    This checks main credentials (gcloud auth login)
    Not application default credentials used by lapdog (gcloud auth application-default login)
    60 second cache (because user can change sign-in)
    """
    return subprocess.run(
        'gcloud config get-value account',
        shell=True,
        stdout=subprocess.PIPE
    ).stdout.decode().strip()

class AuthenticationError(ValueError):
    pass

class InvalidToken(AuthenticationError):
    pass

class AccountMismatch(AuthenticationError):
    pass

class BadDomain(AuthenticationError):
    pass

class LapdogToken(object):
    def __init__(self, account=None):
        """
        Represents a self-refreshing access token for Lapdog APIs, issued by Google.
        Credentials are saved to disk, so each account should only need to login once.
        Access token is valid for 1 hour, but can be refreshed with .refresh()
        """
        self.account = account if account is not None else get_gcloud_account()
        if not self.account.endswith('@broadinstitute.org'):
            raise BadDomain("Non-broad emails are currently unsupported by the Lapdog OAuth system")
        self.token = None
        self.refresh_token = None
        self.ident = None
        self._lock = RLock()
        self._session = None

        path = LapdogToken.path_for_account(self.account)
        try:
            with open(path) as w:
                data = json.load(w)

            self.token = data['access_token']
            self.refresh_token = data['refresh_token']
            self.ident = data['id_token']

        except (FileNotFoundError, json.decoder.JSONDecodeError):
            self.auto_login()

        if not self.valid:
            return self.refresh()

        if self.info['email'] != self.account:
            raise AccountMismatch("{} != {}".format(self.info['email'], self.account))

    @staticmethod
    def path_for_account(account):
        account = sha256(('66e21f0205b97437399244c1095452a0' + getuser() + account).encode()).hexdigest()
        base_path = os.path.expanduser(os.path.join(
            '~',
            '.config',
            'lapdog',
            'auth',
            account[:2],
            account[2:4]
        ))
        if not os.path.exists(base_path):
            os.makedirs(base_path)
        return os.path.join(
            base_path,
            account[4:]
        )

    @staticmethod
    @cached(10)
    def _validate(token, account):
        info = utils.get_token_info(token)
        return (
            'error' not in info
            and 'audience' in info
            and info['audience'] == utils.OAUTH_CLIENT_ID
            and 'expires_in' in info
            and info['expires_in'] > 10 # 10s is important. This way token will not expire during current cache window
            and 'email' in info
            and info['email'] == account
        )

    @property
    def valid(self):
        """
        Checks that the token is valid and not expired
        """
        return LapdogToken._validate(self.token, self.account)


    @property
    def auto_token(self):
        """
        Returns authentication token.
        If token expires in less than 60s, refresh it.
        If token not logged in, raise Authentication Error
        """
        if not (self.valid and self.info['expires_in'] >= 60):
            self.refresh()
        return self.token

    @property
    @cached(10)
    def info(self):
        """
        Get token info
        """
        return utils.get_token_info(self.token)

    def manual_login(self):
        """
        Manually login to the new token.
        Opens a browser tab for authentication, but requires the user manually
        copy-paste the authorization code into the terminal.
        Handles second-step authentication
        """
        pub, priv = self._generate_verifier()
        # refresh w/ redirect
        self._prompt(
            'urn:ietf:wg:oauth:2.0:oob',
            code_challenge=pub
        )
        # Get user to paste code
        self.authenticate(
            input("Paste authentication token: ").strip(),
            redirect_uri='urn:ietf:wg:oauth:2.0:oob',
            code_verifier=priv
        )

    def browser_login(self, uri):
        """
        Login to the new token by passing authorization code to the provided callback uri.
        Opens a browser tab for authentication, but code is transferred automatically.
        Does not handle second-step authentication.
        You must manually fetch the authorization code from the server which handled
        the callback uri, then call .authenticate()
        """
        pub, priv = self._generate_verifier()
        state = base64.urlsafe_b64encode(os.urandom(32)).decode()
        # refresh w/ redirect
        self._prompt(
            uri,
            code_challenge=pub,
            state=state
        )
        return priv, state

    def auto_login(self, port=4201):
        """
        Main entrypoint for token initialization.
        Checks if Lapdog UI is running, then uses that server for authentication.
        Falls back on manual copy-paste authentication
        """
        try:
            fetch_url = 'http://127.0.0.1:{}/api/v1/auth/fetch?state={{}}'.format(port)
            callback_url = 'http://127.0.0.1:{}/api/v1/auth/callback'.format(port)
            response = requests.get(
                fetch_url.format('marco')
            )
            if response.status_code == 200 and response.json() == 'polo':
                priv, state = self.browser_login(
                    callback_url
                )
                print("Waiting for user to accept browser prompt")
                response = requests.get(
                    fetch_url.format(state)
                )
                while response.status_code == 402:
                    time.sleep(1)
                    response = requests.get(
                        fetch_url.format(state)
                    )
                if response.status_code == 200:
                    return self.authenticate(
                        response.json(),
                        callback_url,
                        code_verifier=priv
                    )
                else:
                    raise AuthenticationError("Invalid response from client-side OAuth ({}) : {}".format(response.status_code, response.text))
        except requests.ConnectionError:
            # UI not running, so do manual login
            pass
        self.manual_login()

    def _generate_verifier(self):
        private_key = base64.urlsafe_b64encode(os.urandom(94)).rstrip(b'=')
        return base64.urlsafe_b64encode(sha256(private_key).digest()).decode().rstrip('='), private_key.decode()

    def _prompt(self, redirect_uri, state=None, code_challenge=None, login_hint=None):
        """
        Opens a browser tab for the user to authenticate
        """
        url = 'https://accounts.google.com/o/oauth2/v2/auth?'

        data = {
            'client_id': utils.OAUTH_CLIENT_ID,
            'redirect_uri': redirect_uri,
            'response_type': 'code',
            'scope': 'email profile openid https://www.googleapis.com/auth/devstorage.read_write',
        }

        if state is not None:
            data['state'] = state

        if code_challenge is not None:
            data['code_challenge_method'] = 'S256'
            data['code_challenge'] = code_challenge

        if login_hint is not None:
             data['login_hint'] = login_hint

        if not webbrowser.open_new_tab(
            url + urlencode(data)
        ):
            print(
                crayons.red("Unable to open browser."),
                'Please visit:',
                crayons.normal(url + urlencode(data), bold=True)
            )

    def authenticate(self, code, redirect_uri, code_verifier=None):
        """
        Performs final authentication. Authorization code is exchanged for
        access token and refresh token
        """
        data = {
            'code': code,
            'client_id': utils.OAUTH_CLIENT_ID,
            'redirect_uri': redirect_uri,
            'grant_type': 'authorization_code'
        }

        if code_verifier is not None:
            data['code_verifier'] = code_verifier

        response = requests.post(
            utils.AUTHENTICATION_URL,
            headers={
                'Content-Type': 'application/json'
            },
            json=data
        )

        if response.status_code == 200:
            auth_data = response.json()
            if LapdogToken._validate(auth_data['access_token'], self.account):
                self._save_info(
                    token=auth_data['access_token'],
                    refresh=auth_data['refresh_token'],
                    ident=auth_data['id_token'] if 'id_token' in auth_data else None
                )
            else:
                raise InvalidToken("Unable to obtain oauth token for this account")
        else:
            raise AuthenticationError("Invalid response from server-side OAuth ({}) : {}".format(response.status_code, response.text))


    def refresh(self, t=None):
        """
        Refreshes a token, assuming the client already has a refresh_token
        """
        if t is not None and self.info['expires_in'] > t:
            # If user specifies a time, only refresh if the token expires in the given window
            return
        response = requests.post(
            utils.AUTHENTICATION_URL,
            headers={
                'Content-Type': 'application/json'
            },
            json={
                'refresh_token': self.refresh_token,
                'client_id': utils.OAUTH_CLIENT_ID,
                'grant_type': 'refresh_token'
            }
        )

        if response.status_code == 200:
            auth_data = response.json()
            if LapdogToken._validate(auth_data['access_token'], self.account):
                self._save_info(
                    token=auth_data['access_token'],
                    refresh=self.refresh_token,
                    ident=auth_data['id_token'] if 'id_token' in auth_data else None
                )
            else:
                self.auto_login()
        else:
            raise AuthenticationError("Invalid response from server-side OAuth ({}) : {}".format(response.status_code, response.text))

    def _credentials(self):
        """
        Return valid credentials object
        """
        return Credentials(
            self.auto_token,
            id_token=self.ident,
            scopes=[
                'email',
                'profile',
                'openid',
                'https://www.googleapis.com/auth/devstorage.read_write'
            ]
        )

    @property
    def authorized_session(self):
        """
        Return an active AuthorizedSession for this user
        """
        with self._lock:
            if self._session is None or self._session.credentials.token != self.token or not self.valid:
                self._session = AuthorizedSession(
                    self._credentials()
                )
            return self._session

    def _save_info(self, token, refresh, ident=None):
        info = utils.get_token_info(token)

        if 'error' in info or 'audience' not in info or info['audience'] != utils.OAUTH_CLIENT_ID or 'expires_in' not in info or info['expires_in'] < 10 or 'email' not in info:
            raise InvalidToken("Invalid token {}".format(repr(info)))

        with open(LapdogToken.path_for_account(info['email']), 'w') as w:
            json.dump({
                'access_token': token,
                'refresh_token': refresh,
                'id_token': ident
            }, w)

        if info['email'] != self.account:
            raise AccountMismatch("{} != {}".format(info['email'], self.account))

        self.token = token
        self.refresh_token = refresh
        self.ident = ident



"""
This module is intended to contain code required to patch a given namespace to the current version of lapdog
Lapdog patch will run __project_admin_apply_patch
This function may:
* Deploy new cloud functions to endpoints with updated versions
* Delete insecure old versions
* Modify project roles
* Update iam policy or bindings
* Regenerate signing keys
"""
from .utils import ld_project_for_namespace, __API_VERSION__, generate_default_session, ld_meta_bucket_for_project, update_iam_policy
from . import _deploy, RESOLUTION_URL
from .. import __version__
from ..gateway import resolve_project_for_namespace, CORE_PERMISSIONS, FUNCTIONS_PERMISSIONS, ADMIN_PERMISSIONS, PET_PERMISSIONS, USER_PERMISSIONS, LAPDOG_SERVICES
from ..lapdog import WorkspaceManager
from dalmatian import getblob
import sys
import crayons
import traceback
from firecloud import api as fc
from hashlib import sha512
import requests
import os
import time
from uuid import uuid4

# Notes for creating future patches
# 1) Import a Gateway and use get_endpoint to check for any redacted endpoints
#       Redact any endpoints which can be found
# 2) Always redeploy the current latest version of all endpoints/role definitions/api bindings
# 3) Mark the separate phases (iam, deploy, redact)

# Self update concept:
# 1) Webhook from lapdog github on new tags -> a central cloud endpoint
# 2) Endpoint iterates over all resolutions to trigger update endpoint in each
# 3) update endpoints boot f1-micro instances to checkout tagged commit  and run apply-patch

# Alternatively, write a function for each patch
# When running patches, check that the previous patch has been applied

__REDACTIONS=[
    'register-alpha',
    'register-beta',
    'register-v1',
    'submit-alpha',
    'submit-beta',
    'submit-v1',
    'submit-v2',
    'submit-v3a',
    'abort-alpha',
    'abort-beta',
    'query-alpha',
    'query-beta',
    'quotas-alpha',
    'quotas-beta',
    'signature-alpha',
    'signature-beta',
    'update-v2'
]

__ENDPOINTS__ = {
    'submit': 'create_submission',
    'abort': 'abort_submission',
    'signature': 'check_abort',
    'register': 'register',
    'query': 'query_account',
    'quotas': 'quotas',
    'existence': 'existence',
    'update': 'update'
}

def __project_admin_apply_patch(namespace):
    """
    PATCH SPEC: Beta -> V1
    """
    print("Patch version", __version__)
    print("Patching Namespace:", namespace)
    user_session = generate_default_session()
    print(crayons.normal("Phase 1/7:", bold=True), "Checking resolution status")
    blob = getblob(
        'gs://lapdog-resolutions/' + sha512(namespace.encode()).hexdigest()
    )
    if not blob.exists():
        print("Patching resolution")
        response = user_session.post(
            RESOLUTION_URL,
            headers={"Content-Type": "application/json"},
            json={
                'namespace': namespace,
                'project': ld_project_for_namespace(namespace)
            }
        )
        if response.status_code == 409:
            proj = resolve_project_for_namespace(project_id)
            if proj != custom_lapdog_project:
                raise NameError("A resolution is already in place for this namespace. Please contact GitHub @agraubert")
        elif response.status_code != 200:
            print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
            raise ValueError("Error when generating resolution")
        project = ld_project_for_namespace(namespace)
    else:
        project = blob.download_as_string().decode()
        print(crayons.green("Remote Resolution intact"))
    print("Lapdog Project:", crayons.normal(project, bold=True))
    blob = getblob(
        'gs://{bucket}/resolution'.format(
            bucket = ld_meta_bucket_for_project(project)
        )
    )
    if not blob.exists():
        print("Patching resolution")
        blob.upload_from_string(namespace.encode())
        acl = blob.acl
        acl.all_authenticated().grant_read()
        acl.save()
    else:
        print(crayons.green("Local Resolution intact"))
    functions_account = 'lapdog-functions@{}.iam.gserviceaccount.com'.format(project)
    roles_url = "https://iam.googleapis.com/v1/projects/{project}/roles".format(
        project=project
    )
    print(crayons.normal("Phase 2/7:", bold=True), "Update IAM Policies")
    patch_role(user_session, roles_url, "Core_account", CORE_PERMISSIONS)
    patch_role(user_session, roles_url, "Functions_account", FUNCTIONS_PERMISSIONS)
    patch_role(user_session, roles_url, "Engine_Admin", ADMIN_PERMISSIONS)
    patch_role(user_session, roles_url, "Pet_account", PET_PERMISSIONS)
    patch_role(user_session, roles_url, "Lapdog_user", USER_PERMISSIONS)
    print(crayons.normal("Phase 3/7:", bold=True), "Checking enabled services")
    response = user_session.get(
        "https://serviceusage.googleapis.com/v1/projects/{}/services?filter=state:ENABLED".format(
            project
        )
    )
    if response.status_code != 200:
        raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
    enabled_services = {
        os.path.basename(service['name'])
        for service in response.json()['services']
    }
    required_services = [service for service in LAPDOG_SERVICES if service not in enabled_services]
    if len(required_services):
        print("Enabling additional APIS:", required_services)
        enable_url = "https://serviceusage.googleapis.com/v1/projects/{}/services:batchEnable".format(
            project
        )
        print("POST", enable_url)
        response = user_session.post(
            enable_url,
            json={'serviceIds': required_services}
        )
        if response.status_code != 200:
            raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
    else:
        print(crayons.green("All required services enabled"))
    print(crayons.normal("Phase 4/7:", bold=True), "Checking service accounts")
    response = user_session.get(
        "https://iam.googleapis.com/v1/projects/{}/serviceAccounts".format(project)
    )
    if response.status_code != 200:
        raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
    service_accounts = {
        account['email']
        for account in response.json()['accounts']
    }
    required_acconts = [
        ('lapdog-worker', 'Core_account'),
        ('lapdog-functions', 'Functions_account'),
        ('lapdog-update', 'Engine_Admin')
    ]
    for account, perms in required_acconts:
        email = '{}@{}.iam.gserviceaccount.com'.format(
            account,
            project
        )
        if email not in service_accounts:
            print(crayons.red("Missing account:"), email)
            response = user_session.post(
                'https://iam.googleapis.com/v1/projects/{}/serviceAccounts'.format(project),
                headers={'Content-Type': 'application/json'},
                json={
                    'accountId': account,
                    'serviceAccount': {
                        'displayName': account
                    }
                }
            )
            print("PATCH", 'https://iam.googleapis.com/v1/projects/{}/serviceAccounts'.format(project))
            if response.status_code != 200:
                raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
            update_iam_policy(
                user_session,
                {'serviceAccount:'+email: perms},
                project
            )
    response = user_session.post(
        'https://iam.googleapis.com/v1/projects/{project}/serviceAccounts/lapdog-update@{project}.iam.gserviceaccount.com:setIamPolicy'.format(
            project=project,
        ),
        headers={'Content-Type': 'application/json'},
        json={
            "policy": {
                "bindings": [
                    {
                        "role": "roles/iam.serviceAccountUser",
                        "members": [
                            # Allows the gcloud functions account to set this pet account on comwell servers
                            "serviceAccount:lapdog-functions@{project}.iam.gserviceaccount.com".format(project=project),
                        ]
                    }
                ]
            },
            "updateMask": "bindings"
        }
    )
    if response.status_code != 200:
        print(crayons.red("Warning:", bold=True), "Unable to update lapdog-update service account permissions. The self-update system may not work")
        print("({}) : {}".format(response.status_code, response.text), file=sys.stderr)
    response = user_session.post(
        'https://iam.googleapis.com/v1/projects/{project}/serviceAccounts/lapdog-functions@{project}.iam.gserviceaccount.com:setIamPolicy'.format(
            project=project,
        ),
        headers={'Content-Type': 'application/json'},
        json={
            "policy": {
                "bindings": [
                    {
                        "role": "roles/iam.serviceAccountUser",
                        "members": [
                            # Allows the self-update account to set the functions account on new cloud functions
                            "serviceAccount:lapdog-update@{project}.iam.gserviceaccount.com".format(project=project),
                        ]
                    }
                ]
            },
            "updateMask": "bindings"
        }
    )
    if response.status_code != 200:
        print(crayons.red("Warning:", bold=True), "Unable to update lapdog-update service account permissions. The self-update system may not work")
        print("({}) : {}".format(response.status_code, response.text), file=sys.stderr)
    print(crayons.normal("Phase 5/7:", bold=True), "Checking VPC Configuration")
    blob = getblob('gs://{bucket}/regions'.format(bucket=ld_meta_bucket_for_project(project)))
    regions = ['us-central1']
    try:
        if blob.exists():
            regions = blob.download_as_string().decode().split()
            acl = blob.acl
            acl.all_authenticated().grant_read()
            acl.save()
    except:
        pass
    for region in regions:
        subnet_url = "https://www.googleapis.com/compute/v1/projects/{project}/regions/{region}/subnetworks/default".format(
            project=project,
            region=region
        )
        response = user_session.get(subnet_url)
        if response.status_code != 200:
            raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
        subnet = response.json()
        if not ('privateIpGoogleAccess' in subnet and subnet['privateIpGoogleAccess']):
            print("Patching VPC Configuration in region", region)
            print("POST", subnet_url+'/setPrivateIpGoogleAccess')
            response = user_session.post(
                subnet_url+'/setPrivateIpGoogleAccess',
                headers={
                    'Content-Type': "application/json"
                },
                params={
                    'requestId': str(uuid4())
                },
                json={
                    "privateIpGoogleAccess": True
                }
            )
            if response.status_code >= 400:
                raise ValueError("Unexpected response from Google (%d) : %s" % (response.status_code, response.text))
        else:
            print(crayons.green("VPC Configuration Valid for region "+region))
    print(crayons.normal("Phase 6/7:", bold=True), "Deploy Cloud API Updates")
    response = user_session.get(
        'https://cloudfunctions.googleapis.com/v1/projects/{project}/locations/us-central1/functions'.format(
            project=project
        )
    )
    if response.status_code != 200:
        print("Unable to query existing functions. Redeploying all functions")
        deployments = {
            f:v
            for f,v in __API_VERSION__.items()
            if f not in {'resolve', 'oauth'}
        }
    else:
        functions = {
            function['name'].split('/')[-1]
            for function in response.json()['functions']
            if function['entryPoint'] != 'redacted'
        }
        deployments = {
            f:v
            for f,v in __API_VERSION__.items()
            if (f+'-'+v) not in functions and f not in {'resolve', 'oauth'}
        }
    if len(deployments):
        print("Deploying", len(deployments), "functions")
        for func, ver in deployments.items():
            print("Deploying endpoint", func, ver)
            _deploy(__ENDPOINTS__[func], func, functions_account, project)
    else:
        print(crayons.green("No updates"))
    print(crayons.normal("Phase 7/7:", bold=True), "Redact Insecure Cloud API Endpoints")
    if response.status_code != 200:
        print("Unable to query existing functions. Applying all redactions")
        redactions = __REDACTIONS
    else:
        redactions = [
            function['name'].split('/')[-1]
            for function in response.json()['functions']
            if function['name'].split('/')[-1] in __REDACTIONS
            and function['entryPoint'] != 'redacted'
        ]
    if len(redactions) == 0:
        print(crayons.green("All endpoints secure"))
        return
    print(crayons.normal("%d insecure endpoints detected"%len(redactions), bold=True))
    for redaction in redactions:
        print(crayons.red("Redacting "+redaction))
        endpoint, version = redaction.split('-')
        _deploy('redacted', endpoint, functions_account, project, version)

def patch_role(session, url, title, permissions):
    url = "{}/{}".format(url, title)
    response = session.get(
        url
    )
    if response.status_code == 404:
        print("Redeploying Policy")
        print("POST", url)
        response = session.post(
            url,
            headers={
                'Content-Type': 'application/json'
            },
            json={
                "roleId": title,
                "role": {
                    "title": title,
                    "includedPermissions": permissions,
                    "stage": "GA"
                }
            }
        )
    elif response.status_code == 200:
        perms = {*response.json()['includedPermissions']}
        if len(perms ^ set(permissions)):
            print("Updating Policy")
            print("PATCH", url)
            response = session.patch(
                url,
                headers={
                    'Content-Type': 'application/json'
                },
                json={
                    "title": title,
                    "includedPermissions": permissions,
                    "stage": "GA"
                }
            )
            if response.status_code != 200:
                print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
                raise ValueError("Invalid response from Google API")
        else:
            print(crayons.green("Policy OK"))
    else:
        print("(%d) : %s" % (response.status_code, response.text), file=sys.stderr)
        raise ValueError("Invalid response from Google API")



import os
try:
    from . import utils
except ImportError:
    import sys
    sys.path.append(os.path.dirname(__file__))
    import utils
import traceback

@utils.cors('POST')
def query_account(request):
    logger = utils.CloudLogger().log_request(request)
    try:

        # 1) Validate the token

        token = utils.extract_token(request.headers, None)
        if token is None:
            return (
                {
                    'error': 'Bad Request',
                    'message': 'Token must be provided in header or body'
                },
                400
            )

        token_info = utils.get_token_info(token)
        if 'error' in token_info:
            return (
                {
                    'error': 'Invalid Token',
                    'message': token_info['error_description'] if 'error_description' in token_info else 'Google rejected the client token'
                },
                401
            )

        if not utils.validate_token(token_info):
            return (
                {
                    'error': 'Rejected token',
                    'message': 'Token was valid but did not meet Lapdog security requirements. Token must have email, profile, openid, and devstorage.read_write scopes.'
                    ' Broad users must authenticate via a LapdogToken'
                },
                403
            )

        # 2) Check service account
        default_session = utils.generate_default_session(scopes=['https://www.googleapis.com/auth/cloud-platform'])
        account_email = utils.ld_acct_in_project(token_info['email'])
        response = utils.query_service_account(default_session, account_email)
        if response.status_code >= 400:
            return (
                {
                    'error': 'Unable to query service account',
                    'message': response.text
                },
                400
            )
        if response.json()['email'] != account_email:
            return (
                {
                    'error': 'Service account email did not match expected value',
                    'message': response.json()['email'] + ' != ' + account_email
                },
                400
            )
        return account_email, 200

    except:
        logger.log_exception()
        return (
            {
                'error': 'Unknown Error',
                'message': traceback.format_exc()
            },
            500
        )
